# Copyright (c) 2019-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from logging import getLogger
import math
import itertools
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

# TODO need a more elegant way to do this
N_MAX_POSITIONS = 602  # maximum summary sequence length

DECODER_ONLY_PARAMS = [
    'layer_norm15.%i.weight', 'layer_norm15.%i.bias',
    'encoder_attn.%i.q_lin.weight', 'encoder_attn.%i.q_lin.bias',
    'encoder_attn.%i.k_lin.weight', 'encoder_attn.%i.k_lin.bias',
    'encoder_attn.%i.v_lin.weight', 'encoder_attn.%i.v_lin.bias',
    'encoder_attn.%i.out_lin.weight', 'encoder_attn.%i.out_lin.bias'
]

TRANSFORMER_LAYER_PARAMS = [
    'attentions.%i.q_lin.weight', 'attentions.%i.q_lin.bias',
    'attentions.%i.k_lin.weight', 'attentions.%i.k_lin.bias',
    'attentions.%i.v_lin.weight', 'attentions.%i.v_lin.bias',
    'attentions.%i.out_lin.weight', 'attentions.%i.out_lin.bias',
    'layer_norm1.%i.weight', 'layer_norm1.%i.bias',
    'ffns.%i.lin1.weight', 'ffns.%i.lin1.bias',
    'ffns.%i.lin2.weight', 'ffns.%i.lin2.bias',
    'layer_norm2.%i.weight', 'layer_norm2.%i.bias'
]


logger = getLogger()

def smoothed_softmax_cross_entropy_with_logits(logits, labels, smoothing=0.0):
    if not smoothing:
        loss = F.cross_entropy(logits, labels, reduction='mean')
        return loss

    vocab_size = logits.size(1)
    n = (vocab_size - 1)
    p = 1.0 - smoothing
    q = smoothing / n

    one_hot = torch.randn(1, vocab_size, device=logits.device)
    one_hot.fill_(q)
    soft_targets = one_hot.repeat(labels.view(-1, 1).size(0), 1)
    soft_targets.scatter_(1, labels.view(-1, 1), p)

    log_prb = F.log_softmax(logits, dim=1)
    loss = -(soft_targets * log_prb).sum(dim=1)
    loss = loss.sum() / len(labels)
    return loss

def Embedding(num_embeddings, embedding_dim, padding_idx=None):
    m = nn.Embedding(num_embeddings, embedding_dim, padding_idx=padding_idx)
    nn.init.normal_(m.weight, mean=0, std=embedding_dim ** -0.5)
    if padding_idx is not None:
        nn.init.constant_(m.weight[padding_idx], 0)
    return m


def Linear(in_features, out_features, bias=True):
    m = nn.Linear(in_features, out_features, bias)
    # nn.init.normal_(m.weight, mean=0, std=1)
    # nn.init.xavier_uniform_(m.weight)
    # nn.init.constant_(m.bias, 0.)
    return m


def create_sinusoidal_embeddings(n_pos, dim, out):
    position_enc = np.array([
        [pos / np.power(10000, 2 * (j // 2) / dim) for j in range(dim)]
        for pos in range(n_pos)
    ])
    out[:, 0::2] = torch.FloatTensor(np.sin(position_enc[:, 0::2]))
    out[:, 1::2] = torch.FloatTensor(np.cos(position_enc[:, 1::2]))
    out.detach_()
    out.requires_grad = False


def gelu(x):
    """
    GELU activation
    https://arxiv.org/abs/1606.08415
    https://github.com/huggingface/pytorch-openai-transformer-lm/blob/master/model_pytorch.py#L14
    https://github.com/huggingface/pytorch-pretrained-BERT/blob/master/modeling.py
    """
    # return 0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))
    return 0.5 * x * (1.0 + torch.erf(x / math.sqrt(2.0)))


def get_masks(slen, lengths, causal):
    """
    Generate hidden states mask, and optionally an attention mask.
    """
    assert lengths.max().item() <= slen
    bs = lengths.size(0)
    alen = torch.arange(slen, dtype=torch.long, device=lengths.device)
    mask = alen < lengths[:, None]

    # attention mask is the same as mask, or triangular inferior attention (causal)
    if causal:
        attn_mask = alen[None, None, :].repeat(bs, slen, 1) <= alen[None, :, None]
    else:
        attn_mask = mask

    # sanity check
    assert mask.size() == (bs, slen)
    assert causal is False or attn_mask.size() == (bs, slen, slen)

    return mask, attn_mask

class BinaryOutputLayer(nn.Module):
    """
    0/1 Classification layer (binary classification).
    """
    def __init__(self, params):
        super().__init__()
        self.pad_index = params.pad_index
        dim = params.emb_dim
        self.proj = Linear(dim, 1, bias=True)
        self.proj_act = nn.Sigmoid()
        self.criterion = nn.BCELoss()

    def forward(self, x, y):
        """
        Compute the loss, and optionally the scores.
        """
        scores = self.proj(x)
        scores = self.proj_act(scores)
        y = y.view_as(scores).float()
        loss = self.criterion(scores, y)
        return scores, loss

    def get_scores(self, x):
        """
        Compute scores.
        """
        scores = self.proj(x)
        scores = self.proj_act(scores)
        return scores


class PredLayer(nn.Module):
    """
    Prediction layer (cross_entropy or adaptive_softmax).
    """
    def __init__(self, params):
        super().__init__()
        self.tgt_n_words = params.tgt_n_words
        self.pad_index = params.pad_index
        dim = params.emb_dim
        try:
            self.label_smoothing = params.label_smoothing
        except:
            self.label_smoothing = 0.0

        self.proj = Linear(dim, self.tgt_n_words, bias=True)

    def forward(self, x, y):
        """
        Compute the loss, and optionally the scores.
        """
        assert (y == self.pad_index).sum().item() == 0
        logits = self.proj(x).view(-1, self.tgt_n_words)
        loss = smoothed_softmax_cross_entropy_with_logits(logits, y, self.label_smoothing)

        return logits, loss

    def get_scores(self, x):
        """
        Compute scores.
        """
        return self.proj(x)


class MultiHeadAttention(nn.Module):

    NEW_ID = itertools.count()

    def __init__(self, n_heads, dim, dropout):
        super().__init__()
        self.layer_id = next(MultiHeadAttention.NEW_ID)
        self.dim = dim
        self.n_heads = n_heads
        self.dropout = dropout
        assert self.dim % self.n_heads == 0

        self.q_lin = Linear(dim, dim)
        self.k_lin = Linear(dim, dim)
        self.v_lin = Linear(dim, dim)
        self.out_lin = Linear(dim, dim)

    def forward(self, input, mask, kv=None, cache=None):
        """
        Self-attention (if kv is None) or attention over source sentence (provided by kv).
        """
        # Input is (bs, qlen, dim)
        # Mask is (bs, klen) (non-causal) or (bs, klen, klen)
        bs, qlen, dim = input.size()
        if kv is None:
            klen = qlen if cache is None else cache['slen'] + qlen
        else:
            klen = kv.size(1)
        assert dim == self.dim, 'Dimensions do not match: %s input vs %s configured' % (dim, self.dim)
        n_heads = self.n_heads
        dim_per_head = dim // n_heads
        mask_reshape = (bs, 1, qlen, klen) if mask.dim() == 3 else (bs, 1, 1, klen)

        def shape(x):
            """  projection """
            return x.view(bs, -1, self.n_heads, dim_per_head).transpose(1, 2)

        def unshape(x):
            """  compute context """
            return x.transpose(1, 2).contiguous().view(bs, -1, self.n_heads * dim_per_head)

        q = shape(self.q_lin(input))                                          # (bs, n_heads, qlen, dim_per_head)
        if kv is None:
            k = shape(self.k_lin(input))                                      # (bs, n_heads, qlen, dim_per_head)
            v = shape(self.v_lin(input))                                      # (bs, n_heads, qlen, dim_per_head)
        elif cache is None or self.layer_id not in cache:
            k = v = kv
            k = shape(self.k_lin(k))                                          # (bs, n_heads, qlen, dim_per_head)
            v = shape(self.v_lin(v))                                          # (bs, n_heads, qlen, dim_per_head)

        if cache is not None:
            if self.layer_id in cache:
                if kv is None:
                    k_, v_ = cache[self.layer_id]
                    k = torch.cat([k_, k], dim=2)                             # (bs, n_heads, klen, dim_per_head)
                    v = torch.cat([v_, v], dim=2)                             # (bs, n_heads, klen, dim_per_head)
                else:
                    k, v = cache[self.layer_id]
            cache[self.layer_id] = (k, v)

        q = q / math.sqrt(dim_per_head)                                       # (bs, n_heads, qlen, dim_per_head)
        scores = torch.matmul(q, k.transpose(2, 3))                           # (bs, n_heads, qlen, klen)
        mask = (mask == 0).view(mask_reshape).expand_as(scores)               # (bs, n_heads, qlen, klen)
        scores.masked_fill_(mask, -float('inf'))                              # (bs, n_heads, qlen, klen)

        weights = F.softmax(scores.float(), dim=-1).type_as(scores)           # (bs, n_heads, qlen, klen)
        weights = F.dropout(weights, p=self.dropout, training=self.training)  # (bs, n_heads, qlen, klen)
        context = torch.matmul(weights, v)                                    # (bs, n_heads, qlen, dim_per_head)
        context = unshape(context)                                            # (bs, qlen, dim)

        return self.out_lin(context)


class TransformerFFN(nn.Module):

    def __init__(self, in_dim, dim_hidden, out_dim, dropout, gelu_activation):
        super().__init__()
        self.dropout = dropout
        self.lin1 = Linear(in_dim, dim_hidden)
        self.lin2 = Linear(dim_hidden, out_dim)
        self.act = gelu if gelu_activation else F.relu

    def forward(self, input):
        x = self.lin1(input)
        x = self.act(x)
        x = self.lin2(x)
        x = F.dropout(x, p=self.dropout, training=self.training)
        return x


class TransformerEncoder(nn.Module):

    def __init__(self, params, dico, with_output):
        """
        Transformer model (encoder or decoder).
        """
        super().__init__()

        self.with_output = with_output
        self.with_positional_emb = params.encoder_positional_emb

        self.n_words = params.src_n_words

        self.eos_index = params.eos_index
        self.pad_index = params.pad_index
        self.dico = dico
        assert len(self.dico) == self.n_words

        # model parameters
        self.emb_dim = params.emb_dim // 4  # 128 by default
        self.dim = params.emb_dim       # 512 by default
        self.hidden_dim = self.dim * 4  # 2048 by default
        self.n_heads = params.n_heads   # 8 by default
        self.n_layers = params.enc_n_layers
        self.dropout = params.dropout
        self.attention_dropout = params.attention_dropout
        assert self.dim % self.n_heads == 0, 'transformer dim must be a multiple of n_heads'

        # embeddings
        # TODO remove the hardcode number
        if params.encoder_positional_emb:
            self.position_embeddings = Embedding(800, self.dim)
            if params.sinusoidal_embeddings:
                create_sinusoidal_embeddings(800, self.dim, out=self.position_embeddings.weight)
        self.embeddings = Embedding(self.n_words, self.emb_dim, padding_idx=self.pad_index)
        self.layer_norm_emb = nn.LayerNorm(self.dim, eps=1e-12)

        # transformer layers
        self.attentions = nn.ModuleList()
        self.layer_norm1 = nn.ModuleList()
        self.ffns = nn.ModuleList()
        self.layer_norm2 = nn.ModuleList()

        for _ in range(self.n_layers):
            self.attentions.append(MultiHeadAttention(self.n_heads, self.dim, dropout=self.attention_dropout))
            self.layer_norm1.append(nn.LayerNorm(self.dim, eps=1e-12))
            self.ffns.append(TransformerFFN(self.dim, self.hidden_dim, self.dim, dropout=self.dropout, gelu_activation=params.gelu_activation))
            self.layer_norm2.append(nn.LayerNorm(self.dim, eps=1e-12))

        # output layer
        self.pred_layer = BinaryOutputLayer(params)

    def forward(self, mode, **kwargs):
        """
        Forward function with different forward modes.
        ### Small hack to handle PyTorch distributed.
        """
        if mode == 'fwd':
            return self.fwd(**kwargs)
        elif mode == 'predict':
            return self.predict(**kwargs)
        elif mode == 'score':
            return self.get_scores(**kwargs)
        else:
            raise Exception("Unknown mode: %s" % mode)

    def fwd(self, x1, x2, x3, x4, lengths):
        """
        Inputs:
            `x` LongTensor(slen, bs), containing word indices
            `lengths` LongTensor(bs), containing the length of each sentence
            `causal` Boolean, if True, the attention is only done over previous hidden states
        """
        # lengths = (x != self.pad_index).float().sum(dim=1)
        # mask = x != self.pad_index

        # check inputs
        slen, bs = x1.size()
        assert lengths.size(0) == bs
        assert lengths.max().item() <= slen
        x1 = self.embeddings(x1.transpose(0, 1))
        x2 = self.embeddings(x2.transpose(0, 1))
        x3 = self.embeddings(x3.transpose(0, 1))
        x4 = self.embeddings(x4.transpose(0, 1))

        # generate masks
        mask, attn_mask = get_masks(slen, lengths, causal=False)

        positions = x1.new(slen).long()
        positions = torch.arange(slen, out=positions).unsqueeze(0)

        # embeddings
        tensor = torch.cat((x1, x2, x3, x4), dim=-1)
        if self.with_positional_emb:
            tensor = tensor + self.position_embeddings(positions).expand_as(tensor)
        #tensor = x1 + x2 + x3 + x4
        tensor = self.layer_norm_emb(tensor)
        tensor = F.dropout(tensor, p=self.dropout, training=self.training)
        tensor *= mask.unsqueeze(-1).to(tensor.dtype)

        # transformer layers
        for i in range(self.n_layers):

            # self attention
            attn = self.attentions[i](tensor, attn_mask)
            attn = F.dropout(attn, p=self.dropout, training=self.training)
            tensor = tensor + attn
            tensor = self.layer_norm1[i](tensor)

            # FFN
            tensor = tensor + self.ffns[i](tensor)
            tensor = self.layer_norm2[i](tensor)
            tensor *= mask.unsqueeze(-1).to(tensor.dtype)

        # move back sequence length to dimension 0
        tensor = tensor.transpose(0, 1)

        return tensor

    def predict(self, tensor, pred_mask, y):
        """
        Given the last hidden state, compute word scores and/or the loss.
            `pred_mask` is a ByteTensor of shape (slen, bs), filled with 1 when
                we need to predict a word
            `y` is a LongTensor of shape (pred_mask.sum(),)
        """
        masked_tensor = tensor[pred_mask.unsqueeze(-1).expand_as(tensor)].view(-1, self.dim)
        scores, loss = self.pred_layer(masked_tensor, y)
        return scores, loss

    def get_scores(self, tensor):
        scores = self.pred_layer.get_scores(tensor)
        return scores


class TransformerDecoder(nn.Module):

    def __init__(self, params, dico, with_output):
        """
        Transformer model (encoder or decoder).
        """
        super().__init__()

        # encoder / decoder, output layer
        self.with_output = with_output

        # dictionary / languages
        self.n_words = params.tgt_n_words
            
        self.eos_index = params.eos_index
        self.pad_index = params.pad_index
        self.dico = dico
        print(self.n_words)
        print(len(self.dico))
        assert len(self.dico) == (self.n_words)

        # model parameters
        self.dim = params.emb_dim       # 512 by default
        self.hidden_dim = self.dim * 4  # 2048 by default
        self.n_heads = params.n_heads   # 8 by default
        self.n_layers = params.dec_n_layers
        self.dropout = params.dropout
        self.attention_dropout = params.attention_dropout
        assert self.dim % self.n_heads == 0, 'transformer dim must be a multiple of n_heads'

        # embeddings
        self.position_embeddings = Embedding(N_MAX_POSITIONS, self.dim)
        if params.sinusoidal_embeddings:
            create_sinusoidal_embeddings(N_MAX_POSITIONS, self.dim, out=self.position_embeddings.weight)
        self.embeddings = Embedding(self.n_words, self.dim, padding_idx=self.pad_index)
        self.layer_norm_emb = nn.LayerNorm(self.dim, eps=1e-12)

        # transformer layers
        self.attentions = nn.ModuleList()
        self.layer_norm1 = nn.ModuleList()
        self.ffns = nn.ModuleList()
        self.layer_norm2 = nn.ModuleList()
        self.layer_norm15 = nn.ModuleList()
        self.encoder_attn = nn.ModuleList()

        for _ in range(self.n_layers):
            self.attentions.append(MultiHeadAttention(self.n_heads, self.dim, dropout=self.attention_dropout))
            self.layer_norm1.append(nn.LayerNorm(self.dim, eps=1e-12))
            self.layer_norm15.append(nn.LayerNorm(self.dim, eps=1e-12))
            self.encoder_attn.append(MultiHeadAttention(self.n_heads, self.dim, dropout=self.attention_dropout))
            self.ffns.append(TransformerFFN(self.dim, self.hidden_dim, self.dim, dropout=self.dropout, gelu_activation=params.gelu_activation))
            self.layer_norm2.append(nn.LayerNorm(self.dim, eps=1e-12))

        if self.with_output:
            self.pred_layer = PredLayer(params)
            if params.share_inout_emb:
                self.pred_layer.proj.weight = self.embeddings.weight

    def forward(self, mode, **kwargs):
        """
        Forward function with different forward modes.
        ### Small hack to handle PyTorch distributed.
        """
        if mode == 'fwd':
            return self.fwd(**kwargs)
        elif mode == 'predict':
            return self.predict(**kwargs)
        else:
            raise Exception("Unknown mode: %s" % mode)

    def fwd(self, x, lengths, causal, src_enc=None, src_len=None, positions=None, cache=None):
        """
        Inputs:
            `x` LongTensor(slen, bs), containing word indices
            `lengths` LongTensor(bs), containing the length of each sentence
            `causal` Boolean, if True, the attention is only done over previous hidden states
            `positions` LongTensor(slen, bs), containing word positions
        """
        # lengths = (x != self.pad_index).float().sum(dim=1)
        # mask = x != self.pad_index

        # check inputs
        slen, bs = x.size()
        assert lengths.size(0) == bs
        assert lengths.max().item() <= slen
        x = x.transpose(0, 1)  # batch size as dimension 0
        assert (src_enc is None) == (src_len is None)
        if src_enc is not None:
            assert src_enc.size(0) == bs, "{}!={}".format(src_enc.size(0), bs)

        # generate masks
        mask, attn_mask = get_masks(slen, lengths, causal)
        if src_enc is not None:
            src_mask = torch.arange(src_enc.size(1), dtype=torch.long, device=lengths.device) < src_len[:, None]

        # positions
        if positions is None:
            positions = x.new(slen).long()
            positions = torch.arange(slen, out=positions).unsqueeze(0)
        else:
            assert positions.size() == (slen, bs), positions.size()
            positions = positions.transpose(0, 1)

        # do not recompute cached elements
        if cache is not None:
            _slen = slen - cache['slen']
            x = x[:, -_slen:]
            positions = positions[:, -_slen:]
            mask = mask[:, -_slen:]
            attn_mask = attn_mask[:, -_slen:]

        # embeddings
        tensor = self.embeddings(x)
        tensor = tensor + self.position_embeddings(positions).expand_as(tensor)
        tensor = self.layer_norm_emb(tensor)
        tensor = F.dropout(tensor, p=self.dropout, training=self.training)
        tensor *= mask.unsqueeze(-1).to(tensor.dtype)

        # transformer layers
        for i in range(self.n_layers):

            # self attention
            attn = self.attentions[i](tensor, attn_mask, cache=cache)
            attn = F.dropout(attn, p=self.dropout, training=self.training)
            tensor = tensor + attn
            tensor = self.layer_norm1[i](tensor)

            # encoder attention (for decoder only)
            if src_enc is not None:
                attn = self.encoder_attn[i](tensor, src_mask, kv=src_enc, cache=cache)
                attn = F.dropout(attn, p=self.dropout, training=self.training)
                tensor = tensor + attn
                tensor = self.layer_norm15[i](tensor)

            # FFN
            tensor = tensor + self.ffns[i](tensor)
            tensor = self.layer_norm2[i](tensor)
            tensor *= mask.unsqueeze(-1).to(tensor.dtype)

        # update cache length
        if cache is not None:
            cache['slen'] += tensor.size(1)

        # move back sequence length to dimension 0
        tensor = tensor.transpose(0, 1)

        return tensor

    def predict(self, tensor, pred_mask, y):
        """
        Given the last hidden state, compute word scores and/or the loss.
            `pred_mask` is a ByteTensor of shape (slen, bs), filled with 1 when
                we need to predict a word
            `y` is a LongTensor of shape (pred_mask.sum(),)
        """
        masked_tensor = tensor[pred_mask.unsqueeze(-1).expand_as(tensor)].view(-1, self.dim)
        scores, loss = self.pred_layer(masked_tensor, y)
        return scores, loss

    def generate(self, src_enc, src_len, max_len=200, sample_temperature=None, vocab_mask=None):
        """
        Decode a sentence given initial start.
        `x`:
            - LongTensor(bs, slen)
                <EOS> W1 W2 W3 <EOS> <PAD>
                <EOS> W1 W2 W3   W4  <EOS>
        `lengths`:
            - LongTensor(bs) [5, 6]
        `positions`:
            - False, for regular "arange" positions (LM)
            - True, to reset positions from the new generation (MT)
        """

        # input batch
        bs = len(src_len)
        assert src_enc.size(0) == bs

        # generated sentences
        generated = src_len.new(max_len, bs)  # upcoming output
        generated.fill_(self.pad_index)       # fill upcoming ouput with <PAD>
        generated[0].fill_(self.eos_index)    # we use <EOS> for <BOS> everywhere

        # positions
        positions = src_len.new(max_len).long()
        positions = torch.arange(max_len, out=positions).unsqueeze(1).expand(max_len, bs)

        # current position / max lengths / length of generated sentences / unfinished sentences
        cur_len = 1
        gen_len = src_len.clone().fill_(1)
        unfinished_sents = src_len.clone().fill_(1)

        # cache compute states
        cache = {'slen': 0}

        while cur_len < max_len:

            # compute word scores
            tensor = self.forward(
                'fwd',
                x=generated[:cur_len],
                lengths=gen_len,
                positions=positions[:cur_len],
                causal=True,
                src_enc=src_enc,
                src_len=src_len,
                cache=cache
            )
            assert tensor.size() == (1, bs, self.dim)
            tensor = tensor.data[-1, :, :]               # (bs, dim)
            scores = self.pred_layer.get_scores(tensor)  # (bs, n_words)


            # select next words: sample or greedy
            if sample_temperature is None:
                next_words = torch.topk(scores, 1)[1].squeeze(1)
            else:
                next_words = torch.multinomial(F.softmax(scores / sample_temperature, dim=1), 1).squeeze(1)
            assert next_words.size() == (bs,)

            # update generations / lengths / finished sentences / current length
            generated[cur_len] = next_words * unfinished_sents + self.pad_index * (1 - unfinished_sents)
            gen_len.add_(unfinished_sents)
            unfinished_sents.mul_(next_words.ne(self.eos_index).long())
            cur_len = cur_len + 1

            # stop when there is a </s> in each sentence, or if we exceed the maximul length
            if unfinished_sents.max() == 0:
                break

        # add <EOS> to unfinished sentences
        if cur_len == max_len:
            generated[-1].masked_fill_(unfinished_sents.byte(), self.eos_index)

        # sanity check
        assert (generated == self.eos_index).sum() == 2 * bs

        return generated[:cur_len], gen_len

    def generate_beam(self, src_enc, src_len, beam_size, length_penalty, early_stopping, max_len=200):
        """
        Decode a sentence given initial start.
        `x`:
            - LongTensor(bs, slen)
                <EOS> W1 W2 W3 <EOS> <PAD>
                <EOS> W1 W2 W3   W4  <EOS>
        `lengths`:
            - LongTensor(bs) [5, 6]
        `positions`:
            - False, for regular "arange" positions (LM)
            - True, to reset positions from the new generation (MT)
        """

        # check inputs
        assert src_enc.size(0) == src_len.size(0)
        assert beam_size >= 1

        # batch size / number of words
        bs = len(src_len)
        n_words = self.n_words

        # expand to beam size the source latent representations / source lengths
        src_enc = src_enc.unsqueeze(1).expand((bs, beam_size) + src_enc.shape[1:]).contiguous().view((bs * beam_size,) + src_enc.shape[1:])
        src_len = src_len.unsqueeze(1).expand(bs, beam_size).contiguous().view(-1)

        # generated sentences (batch with beam current hypotheses)
        generated = src_len.new(max_len, bs * beam_size)  # upcoming output
        generated.fill_(self.pad_index)                   # fill upcoming ouput with <PAD>
        generated[0].fill_(self.eos_index)                # we use <EOS> for <BOS> everywhere

        # generated hypotheses
        generated_hyps = [BeamHypotheses(beam_size, max_len, length_penalty, early_stopping) for _ in range(bs)]

        # positions
        positions = src_len.new(max_len).long()
        positions = torch.arange(max_len, out=positions).unsqueeze(1).expand_as(generated)

        # scores for each sentence in the beam
        beam_scores = src_enc.new(bs, beam_size).fill_(0)
        beam_scores[:, 1:] = -1e9
        beam_scores = beam_scores.view(-1)

        # current position
        cur_len = 1

        # cache compute states
        cache = {'slen': 0}

        # done sentences
        done = [False for _ in range(bs)]

        while cur_len < max_len:

            # compute word scores
            tensor = self.forward(
                'fwd',
                x=generated[:cur_len],
                lengths=src_len.new(bs * beam_size).fill_(cur_len),
                positions=positions[:cur_len],
                causal=True,
                src_enc=src_enc,
                src_len=src_len,
                cache=cache
            )
            assert tensor.size() == (1, bs * beam_size, self.dim)
            tensor = tensor.data[-1, :, :]               # (bs * beam_size, dim)
            #print(tensor)
            scores = self.pred_layer.get_scores(tensor)  # (bs * beam_size, n_words)
            scores = F.log_softmax(scores, dim=-1)       # (bs * beam_size, n_words)
            assert scores.size() == (bs * beam_size, n_words)

            # select next words with scores
            #scores of n top sentences where n = beam_size
            _scores = scores + beam_scores[:, None].expand_as(scores)  # (bs * beam_size, n_words)
            #print(f'1 {_scores}')
            #top score?
            _scores = _scores.view(bs, beam_size * n_words)            # (bs, beam_size * n_words)
            #print(f'2 {_scores}')
            next_scores, next_words = torch.topk(_scores, 2 * beam_size, dim=1, largest=True, sorted=True)
            #test = []
            #for idx, value in zip(next_words[0], next_scores[0]):
                # get beam and word IDs
                #beam_id = idx // n_words
                #word_id = idx % n_words
                #test.append(self.dico[word_id.item()])
                #print(next_score, next_word)
                #if next_word in removedDict:
                    #print(next_word)
            #print(f'next words: {test}')
            assert next_scores.size() == next_words.size() == (bs, 2 * beam_size)

            # next batch beam content
            # list of (bs * beam_size) tuple(next hypothesis score, next word, current position in the batch)
            next_batch_beam = []

            # for each sentence
            for sent_id in range(bs):

                # if we are done with this sentence
                done[sent_id] = done[sent_id] or generated_hyps[sent_id].is_done(next_scores[sent_id].max().item())
                if done[sent_id]:
                    next_batch_beam.extend([(0, self.pad_index, 0)] * beam_size)  # pad the batch
                    continue

                # next sentence beam content
                next_sent_beam = []

                # next words for this sentence
                for idx, value in zip(next_words[sent_id], next_scores[sent_id]):

                    # get beam and word IDs
                    beam_id = idx // n_words
                    word_id = idx % n_words
                    #print(, value)
                    #print('loop')

                    # end of sentence, or next word
                    if word_id == self.eos_index or cur_len + 1 == max_len:
                        generated_hyps[sent_id].add(generated[:cur_len, sent_id * beam_size + beam_id].clone(), value.item())
                        #print(f'hyp: {generated_hyps}')
                    else:
                        next_sent_beam.append((value, word_id, sent_id * beam_size + beam_id))
                        #if word_id.item() in self.dico.id2word.keys():
                            #if 'template' in self.dico.id2word[word_id.item()]:
                                #if word_id.item() not in [ids[1] for ids in next_sent_beam]:
                                    #print(self.dico.id2word[word_id.item()])
                                    #next_sent_beam.append((value, word_id, sent_id * beam_size + beam_id))
                                #else:
                                #    print(f'{self.dico.id2word[word_id.item()]} : already in : {[self.dico.id2word[ids[1].item()] for ids in next_sent_beam]}')
                            #else:

                        #else:
                            #print(f'bad {word_id.item()}')

                    # the beam for next step is full
                    if len(next_sent_beam) == beam_size:
                        break
                """
                for i in range(len(next_sent_beam)):
                    #print(self.dico[word_id.item()])
                    if self.dico[word_id.item()] in removedDict:
                        print(next_sent_beam)
                        #if predicted template is not valid, pop the tuple and append it to the end
                        badBeam = next_sent_beam[i].pop()
                        next_sent_beam.append(badBeam)
                        print(next_sent_beam)
                        """
                # update next beam content
                assert len(next_sent_beam) == 0 if cur_len + 1 == max_len else beam_size
                if len(next_sent_beam) == 0:
                    next_sent_beam = [(0, self.pad_index, 0)] * beam_size  # pad the batch
                next_batch_beam.extend(next_sent_beam)
                assert len(next_batch_beam) == beam_size * (sent_id + 1)

            # sanity check / prepare next batch
            assert len(next_batch_beam) == bs * beam_size
            beam_scores = beam_scores.new([x[0] for x in next_batch_beam])
            beam_words = generated.new([x[1] for x in next_batch_beam])
            beam_idx = src_len.new([x[2] for x in next_batch_beam])

            #decoded = [self.dico[x.item()] for x in beam_words]
            #print(f'decoded: {decoded}')

            # re-order batch and internal states
            generated = generated[:, beam_idx]
            generated[cur_len] = beam_words
            #print(generated)
            #print(generated[cur_len])
            for k in cache.keys():
                if k != 'slen':
                    cache[k] = (cache[k][0][beam_idx], cache[k][1][beam_idx])

            # update current length
            cur_len = cur_len + 1

            # stop when we are done with each sentence
            if all(done):
                break

        # visualize hypotheses
        # print([len(x) for x in generated_hyps], cur_len)
        # globals().update( locals() );
        # !import code; code.interact(local=vars())
        #for ii in range(bs):
        #    for ss, ww in sorted(generated_hyps[ii].hyp, key=lambda x: x[0], reverse=True):
        #        print("%.3f " % ss + " ".join(self.dico[x] for x in ww.tolist()))
        #        print("")

        # select the best hypotheses
        tgt_len = src_len.new(bs)
        best = []

        for i, hypotheses in enumerate(generated_hyps):
            best_hyp = max(hypotheses.hyp, key=lambda x: x[0])[1]
            tgt_len[i] = len(best_hyp) + 1  # +1 for the <EOS> symbol
            best.append(best_hyp)

        # generate target batch
        decoded = src_len.new(tgt_len.max().item(), bs).fill_(self.pad_index)
        for i, hypo in enumerate(best):
            decoded[:tgt_len[i] - 1, i] = hypo
            decoded[tgt_len[i] - 1, i] = self.eos_index

        # sanity check
        assert (decoded == self.eos_index).sum() == 2 * bs

        return decoded, tgt_len

    def generate_slot(self, x, xlen, y_type, src_enc, src_len, sample_temperature=None):
        # input batch
        bs = len(src_len)
        assert src_enc.size(0) == bs, "{}!={}".format(src_enc.size(0), bs)
        max_len = xlen.max()

        indices = (x==4).nonzero()
        pred_indices, batch_indices = torch.split(indices, 1, 1)

        # generated sentences
        generated = x.clone()

        # positions
        positions = torch.arange(max_len, dtype=torch.long, device=src_len.device)
        positions = positions.unsqueeze(1).expand(max_len, bs)

        # current position / max lengths / length of generated sentences / unfinished sentences
        cur_idx = 0
        max_idx = pred_indices.size(0)

        while cur_idx < max_idx:
            cur_len = pred_indices[cur_idx].item()
            batch_inx = batch_indices[cur_idx].item()

            cur_type = y_type[cur_len,batch_inx]
            # compute word scores
            tensor = self.forward(
                'fwd',
                x=generated[:cur_len, batch_inx][:,None],
                lengths=pred_indices[cur_idx],
                positions=positions[:cur_len, batch_inx][:,None],
                causal=True,
                src_enc=src_enc[batch_inx][None],
                src_len=src_len[batch_inx][None],
                cache=None
            )
            #assert tensor.size() == (1, bs, self.dim)
            tensor = tensor.data[-1, :, :]               # (bs, dim)
            tensor_type_embedding = self.prediction_type_embeddings(cur_type)
            tensor = torch.cat((tensor, tensor_type_embedding.unsqueeze(0)), dim=-1)
            tensor = self.prediction_type_pooler(tensor)
            scores = self.pred_layer.get_scores(tensor)  # (bs, n_words)


            # select next words: sample or greedy
            if sample_temperature is None:
                # take top-2 to avoid generating <eos>
                next_words = torch.topk(scores, 2)[1].squeeze()
            else:
                next_words = torch.multinomial(F.softmax(scores / sample_temperature, dim=1), 1).squeeze(1)
            assert next_words.size() == (2,), next_words.size()

            # update generations / lengths / finished sentences / current length
            # generated[cur_len, batch_inx] = next_words
            generated[cur_len, batch_inx] = next_words[0] if next_words[0] != self.eos_index else next_words[1]
            cur_idx = cur_idx + 1

        # sanity check
        assert (generated == self.eos_index).sum() == 2 * bs, (generated == self.eos_index).sum()

        return generated

    def generate_slot_beam(self, x, xlen, y_type, src_enc, src_len, beam_size=2):
        # input batch
        bs = len(src_len)
        n_words = self.n_words
        assert src_enc.size(0) == bs, "{}!={}".format(src_enc.size(0), bs)
        max_len = xlen.max()

        # expand to beam size the source latent representations / source lengths
        src_enc = src_enc.unsqueeze(1).expand((bs, beam_size) + src_enc.shape[1:]).contiguous()
        src_len = src_len.unsqueeze(1).expand(bs, beam_size).contiguous()
        
        result = []
        for batch_idx in range(bs):
            x_i = x[:,batch_idx]
            pred_indices = (x_i==4).nonzero()

            src_enc_sub = src_enc[batch_idx]
            src_len_sub = src_len[batch_idx]

            # generated sentences
            generated = x_i.clone()
            generated = generated.unsqueeze(1).expand(max_len, beam_size)

            positions = torch.arange(max_len, dtype=torch.long, device=src_len.device)
            positions = positions.unsqueeze(1).expand(max_len, beam_size)

            # scores for each sentence in the beam
            beam_scores = src_enc.new(1, beam_size).fill_(0)
            beam_scores[:, 1:] = -1e9
            beam_scores = beam_scores.view(-1)

            cur_len = 1
            cache = {'slen': 0}

            while cur_len < max_len:
                cur_type = y_type[cur_len, batch_idx]
                # compute word scores
                tensor = self.forward(
                    'fwd',
                    x=generated[:cur_len],
                    lengths=src_len.new(beam_size).fill_(cur_len),
                    positions=positions[:cur_len],
                    causal=True,
                    src_enc=src_enc_sub,
                    src_len=src_len_sub,
                    cache=cache
                )
                assert tensor.size() == (1, beam_size, self.dim)
                tensor = tensor.data[-1, :, :]               # (beam_size, dim)
                if cur_type.item() == 0:
                    tgt_id = x_i[cur_len].item()
                    scores = self.pred_layer.get_scores(tensor)  # (beam_size, n_words)
                    scores = scores[:, tgt_id] # get the target token score
                    scores = F.log_softmax(scores, dim=-1)       # (beam_size)
                    assert scores.size() == (beam_size,)
                    
                    # here we don't need to sort the beam order
                    beam_scores = beam_scores + scores
                else:
                    tensor_type = self.prediction_type_embeddings(cur_type)
                    tensor_type = tensor_type.unsqueeze(0).expand_as(tensor)
                    tensor = torch.cat((tensor, tensor_type), dim=-1)
                    tensor = self.prediction_type_pooler(tensor)
                    scores = self.pred_layer.get_scores(tensor)  # (beam_size, n_words)
                    scores = F.log_softmax(scores, dim=-1)       # (beam_size, n_words)
                    assert scores.size() == (beam_size, n_words)

                    # select next words with scores
                    _scores = scores + beam_scores[:, None].expand_as(scores)  # (beam_size, n_words)
                    _scores = _scores.view(1, beam_size * n_words)             # (1, beam_size * n_words)

                    next_scores, next_words = torch.topk(_scores, 2 * beam_size, dim=1, largest=True, sorted=True)
                    assert next_scores.size() == next_words.size() == (1, 2 * beam_size)

                    # next sentence beam content
                    next_sent_beam = []

                    for idx, value in zip(next_words[0], next_scores[0]):
                        beam_id = idx // n_words
                        word_id = idx % n_words

                        # end of sentence, or next word
                        if word_id != self.eos_index:
                            next_sent_beam.append((value, word_id, beam_id))

                        # the beam for next step is full
                        if len(next_sent_beam) == beam_size:
                            break

                    beam_scores = beam_scores.new([x[0] for x in next_sent_beam])
                    beam_words  = generated.new([x[1] for x in next_sent_beam])
                    beam_idx = src_len.new([x[2] for x in next_sent_beam])

                    generated = generated[:, beam_idx]
                    generated[cur_len] = beam_words
            
                    for k in cache.keys():
                        if k != 'slen':
                            cache[k] = (cache[k][0][beam_idx], cache[k][1][beam_idx])
                cur_len += 1
            # sanity check
            assert (generated == self.eos_index).sum() == 2 * beam_size, (generated == self.eos_index).sum()
            max_score, index = beam_scores.max(0)
            result.append(generated[:,index].unsqueeze(1))

        result = torch.cat(result, dim=1)
        assert result.size() == x.size()

        return result


class BeamHypotheses(object):

    def __init__(self, n_hyp, max_len, length_penalty, early_stopping):
        """
        Initialize n-best list of hypotheses.
        """
        self.max_len = max_len - 1  # ignoring <BOS>
        self.length_penalty = length_penalty
        self.early_stopping = early_stopping
        self.n_hyp = n_hyp
        self.hyp = []
        self.worst_score = 1e9

    def __len__(self):
        """
        Number of hypotheses in the list.
        """
        return len(self.hyp)

    def add(self, hyp, sum_logprobs):
        """
        Add a new hypothesis to the list.
        """
        #print(self.removedDict)
        score = sum_logprobs / len(hyp) ** self.length_penalty
        if len(self) < self.n_hyp or score > self.worst_score:
            #if hyp not in self.removedDict:
                #print(score, hyp)
            self.hyp.append((score, hyp))
            if len(self) > self.n_hyp:
                sorted_scores = sorted([(s, idx) for idx, (s, _) in enumerate(self.hyp)])
                del self.hyp[sorted_scores[0][1]]
                self.worst_score = sorted_scores[1][0]
            else:
                self.worst_score = min(score, self.worst_score)

    def is_done(self, best_sum_logprobs):
        """
        If there are enough hypotheses and that none of the hypotheses being generated
        can become better than the worst one in the heap, then we are done with this sentence.
        """
        if len(self) < self.n_hyp:
            return False
        elif self.early_stopping:
            return True
        else:
            return self.worst_score >= best_sum_logprobs / self.max_len ** self.length_penalty

